@@ -312,10 +312,13 @@ def run_import(address, id, verbose, dry_run, working_dir, no_lock, terragrunt_a
     help='Show output as json',
     show_default=True
     )
+@click.option('--parallel-execution', '-p', is_flag=True, default=False,
+    help='Whether or not to use parallel execution',
+    )
 @click.argument('terragrunt-args', nargs=-1, type=click.UNPROCESSED)
 @click.version_option(version=__version__)
 def analyze(verbose, dry_run, exclude_external_dependencies, working_dir, 
-    start_at_step, out, terragrunt_args):
+    start_at_step, out, parallel_execution, terragrunt_args):
     """ Analyzes the plan files """
 
     check_latest_version(verbose)
@@ -327,6 +330,7 @@ def analyze(verbose, dry_run, exclude_external_dependencies, working_dir,
         working_dir=working_dir,
         start_at_step=start_at_step,
         out=out,
+        parallel_execution=parallel_execution,
         terragrunt_args=terragrunt_args,
     )
 
@@ -479,7 +483,7 @@ def sync(
     source_domain, source_stage, target_domain, target_stage, module, auto_approve,
     verbose, dry_run, clean_up, include_lock_file, working_dir
     ):
-    """ Deploys the terragrunt config files from one stage to another (and possibly to a different domain) """
+    """ Syncs the terragrunt config files from one stage to another (and possibly to a different domain) """
 
     check_latest_version(verbose)
 
