@@ -21,6 +21,9 @@ import re
 import tempfile
 import json
 import yaml
+import threading
+import queue
+import multiprocessing
 import click
 import hcl
 import networkx as nx
@@ -196,30 +199,50 @@ class TgWrap():
 
         return error, skip
 
-    def _run_graph_step(self, command, working_dir, module, collect_output_file, dry_run, progress):
+    def _run_graph_step(self, command, working_dir, module, collect_output_file, dry_run, progress, 
+        output_queue=None, semaphore=None):
         """ Runs a step in the graph """
 
+        MODULE_IDENTIFIER=f'{module}{self.SEPARATOR}'
+
         stop_processing = False
-
-        # if we have a specific working dir, and the dir is relative, combine the two
-        if working_dir and not os.path.isabs(module):
-            module = os.path.join(os.path.abspath(working_dir), module)
-            self.printer.verbose(f'Executing in directory: {working_dir}')
-
-        self.printer.header(
-            f'\n\nStart processing module: {module} ({progress})\n\n',
-            print_line_before=True,
-            )
+        error = False
+        skip = False
+        output = None
+        error_msg = None
 
         try:
+            # if we are in multi-threading mode, acquire a semaphore
+            if semaphore:
+                semaphore.acquire()
+
+            # if we have a specific working dir, and the dir is relative, combine the two
+            if working_dir and not os.path.isabs(module):
+                module = os.path.join(os.path.abspath(working_dir), module)
+                self.printer.verbose(f'Executing in directory: {working_dir}')
+
+            self.printer.header(
+                f'\n\nStart processing module: {module} ({progress})\n\n',
+                print_line_before=True,
+                )
+
             if dry_run:
                 self.printer.warning(
                     'In dry run mode, no real actions are executed!!'
                     )
             else:
                 if collect_output_file:
-                    collect_output_file.write(f'{module}{self.SEPARATOR}')
+                    self.printer.verbose('Use an output file for output collection')
+
+                    # module identifier needs to be pre-written to output file, so that the output can be appended to it
+                    collect_output_file.write(MODULE_IDENTIFIER)
                     collect_output_file.flush()
+                elif output_queue:
+                    self.printer.verbose('Use an output queue for output collection')
+
+                    # no output file, but data must be written to queue
+                    # so we need to capture the output
+                    collect_output_file = subprocess.PIPE
 
                 messages = ""
 
@@ -233,7 +256,13 @@ class TgWrap():
                         stdout=collect_output_file if collect_output_file else sys.stdout,
                         stderr=f,
                     )
-                    self.printer.verbose(rc)
+
+                    self.printer.verbose(f'arguments: {rc.args}')
+                    self.printer.verbose(f'returncode: {rc.returncode}')
+                    try:
+                        self.printer.verbose(f'stdout: {rc.stdout[:200]}')
+                    except Exception:
+                        pass
 
                 with open(f.name, 'r') as f:
                     messages = f.read()
@@ -244,35 +273,55 @@ class TgWrap():
                     )
                 
                 # if we have a skipped module, and are collecting output, make sure we end up on a new line
-                if skip and collect_output_file:
+
+                output = rc.stdout if rc.stdout else '\n'
+
+                if skip:
                     self.printer.verbose("Module is skipped")
-                    collect_output_file.write('\n')
 
                 if error:
                     raise Exception(
                         f'An error situation detected while processing the terragrunt dependencies graph in directory {module}'
-                        )
+                    )
                 else:
                     self.printer.success(
                         f'Directory {module} processed successfully',
                     )
 
         except FileNotFoundError:
-            self.printer.warning(f'Directory {module} not found, continue')
+            error_msg = f'Directory {module} not found, continue'
+            self.printer.warning(error_msg)
         except Exception as e:
-            self.printer.error(f"Error occurred:\n{str(e)}")
+            error_msg = f"Error occurred:\n{str(e)}"
+            self.printer.error(error_msg)
             self.printer.error("Full stack:", print_line_before=True)
             self.printer.normal(messages, print_line_after=True)
             self.printer.normal(f"Directory {module} failed!")
 
             stop_processing = True
         finally:
+            if error_msg:
+                output = f'{"exception": "{error_msg}"}'
+
+            try:
+                # communicate the results if desired
+                if output_queue:
+                    output_queue.put(f'{MODULE_IDENTIFIER}{output}')
+                elif collect_output_file and (skip or error):
+                    collect_output_file.write(output)
+                    collect_output_file.flush()
+            except Exception as e:
+                self.printer.error(f'Error writing the results: {e}')
+
+            if semaphore:
+                semaphore.release()
+
             os.remove(f.name)
 
         return stop_processing
 
     def _run_di_graph(
-        self, command, exclude_external_dependencies, start_at_step, dry_run,
+        self, command, exclude_external_dependencies, start_at_step, dry_run, parallel_execution,
         ask_for_confirmation=False, collect_output_file=None,
         working_dir=None,
         ):
@@ -301,23 +350,51 @@ class TgWrap():
             if response.lower() != "y":
                 sys.exit(1)
 
+        # We only support multi-threading with 'show'
+        nbr_of_threads = multiprocessing.cpu_count() if parallel_execution and 'show' in command.lower() else 1
+
+        if parallel_execution:
+            self.printer.warning(f'We are in EXPERIMENTAL multi-threading mode using {nbr_of_threads} threads!')
+            q = queue.Queue()
+            semaphore = threading.Semaphore(nbr_of_threads)
+            threads = []
+
+        counter = 0
         nbr_of_groups = len(groups)
         for idx, group in enumerate(groups):
             group_nbr=idx+1
             self.printer.header(f'Group {group_nbr}')
             self.printer.normal(group)
 
-            if command:
-                nbr_of_modules = len(group)
-                for idx2, module in enumerate(group):
-                    module_nbr=idx2+1
-                    progress = f'module {module_nbr} (of {nbr_of_modules}) of group {group_nbr} (of {nbr_of_groups})'
+            nbr_of_modules = len(group)
+            for idx2, module in enumerate(group):
+                counter += 1
+                module_nbr=idx2+1
+                progress = f'module {module_nbr} (of {nbr_of_modules}) of group {group_nbr} (of {nbr_of_groups})'
 
-                    step_nbr = group_nbr + module_nbr/10
-                    if step_nbr < start_at_step:
-                        self.printer.normal(f'Skip step {step_nbr}, start at {start_at_step}')
-                        continue
+                step_nbr = group_nbr + module_nbr/10
+                if step_nbr < start_at_step:
+                    self.printer.normal(f'Skip step {step_nbr}, start at {start_at_step}')
+                    continue
 
+                if parallel_execution:
+                    self.printer.verbose(f'Start thread #{counter} for step {step_nbr}')
+                    t = threading.Thread(
+                        target=self._run_graph_step,
+                        kwargs={
+                            "command": command,
+                            "working_dir": working_dir,
+                            "module": module,
+                            "collect_output_file": None, # in parallel mode we can't write directly to the output file
+                            "dry_run": dry_run,
+                            "progress": progress,
+                            "output_queue": q,
+                            "semaphore": semaphore,
+                            }
+                        )
+                    t.start()
+                    threads.append(t)
+                else:
                     stop_processing = self._run_graph_step(
                         command=command,
                         working_dir=working_dir,
@@ -333,6 +410,17 @@ class TgWrap():
                             f"After you've fixed the problem, you can continue where you were left by adding '--start-at-step {group_nbr}.{module_nbr}'."
                             )
                         sys.exit(1)
+
+        if parallel_execution:
+            # now wait until the threads are done and collect the output
+            # todo: how to implement something as stop_processing like in regular execution?
+            total_counter = counter
+            counter = 0
+            for t in threads:
+                counter += 1
+                self.printer.verbose(f'Wait for thread #{counter} (of {total_counter}) to finish')
+                t.join()
+                collect_output_file.write(q.get())
         
         self.printer.verbose(f'Executed {group_nbr} groups and {step_nbr} steps')
 
@@ -523,7 +611,7 @@ class TgWrap():
             self.printer.verbose(rc)
 
     def analyze(self, dry_run, exclude_external_dependencies, working_dir,
-        start_at_step, out, terragrunt_args):
+        start_at_step, out, parallel_execution, terragrunt_args):
         """ Analyzes the plan files """
 
         self.printer.verbose("Attempting to 'analyze'")
@@ -574,13 +662,17 @@ class TgWrap():
                     working_dir=working_dir,
                     start_at_step=start_at_step,
                     ask_for_confirmation=False,
+                    parallel_execution=parallel_execution,
                 )
 
             with open(f.name, 'r') as f:
                 for line in f:
                     split_line = line.split(self.SEPARATOR)
                     module = split_line[0]
-                    plan_file = split_line[1]
+                    try:
+                        plan_file = split_line[1]
+                    except IndexError:
+                        self.printer.warning(f'Could not determine planfile: {line[:100]}')
 
                     changes[module] = {
                         "deletions" : [],
@@ -593,7 +685,10 @@ class TgWrap():
                         if len(plan_file) > 1:
                             data = json.loads(plan_file)
 
-                            if 'resource_changes' in data and len(data['resource_changes']) > 0:
+                            # do we have an exception logged?
+                            if 'exception' in data:
+                                raise Exception(data['exception'])
+                            elif 'resource_changes' in data and len(data['resource_changes']) > 0:
                                 for rc in data['resource_changes']:
                                     # check if we do have actual changes
                                     actions = rc['change']['actions']
@@ -672,7 +767,7 @@ class TgWrap():
         self, source_stage, target_stage, source_domain, target_domain, module,
         auto_approve, dry_run, clean_up, include_lock_file, working_dir, 
         ):
-        """ Deploys the terragrunt config files from one stage to another (and possibly to a different domain) """
+        """ Syncs the terragrunt config files from one stage to another (and possibly to a different domain) """
     
         if target_domain and not source_domain:
             raise Exception("Providing a target domain while omitting the source domain is not supported!")
@@ -845,8 +940,9 @@ class TgWrap():
                     clean_up=True,
                 )
 
-            with open(os.path.join(target_dir, 'version.hcl'), 'w') as f:
-                f.write(f"""
+            if not dry_run:
+                with open(os.path.join(target_dir, 'version.hcl'), 'w') as f:
+                    f.write(f"""
 locals {{
     version_tag="{version_tag}"
 }}
